$enable-rounded: false;
@import "node_modules/bootstrap/scss/bootstrap";


body,
#root {
    height: 100vh;
    width: 100vw;
}

[data-bs-theme="dark"] {
    --bs-body-bg: #1a1a1a;
    --bs-border-color: #2d2d2d;

    #sidebar {
        background-color: #1a1c1e;
    }

    .nav-pills {
        --bs-nav-pills-link-active-bg: #2d2d2d;
        --bs-nav-pills-link-active-color: #fff;
    }
}

a {
    text-decoration: none;
}